import numpy as np
from scipy import fftpack
from scipy.optimize import fsolve, brentq, minimize
from getdist.convolve import dct2d
import logging
import warnings

"""
Code to find optimal bandwidths for basic kernel density estimators in 1 and 2D
Adapted from Matlab code by Zdravko Botev
Extended to include correlation estimation and numerical AMISE minimization in 2D case
Antony Lewis, 2015-03

Original code notice:

Copyright (c) 2007, Zdravko Botev
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the distribution
    * Neither the name of the The University of Queensland nor the names
      of its contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

rootpi = np.sqrt(np.pi)
pisquared = np.pi ** 2

_kde_lmax_bandwidth = 7
_kde_consts_1d = np.array([(1 + 0.5 ** (j + 0.5)) / 3 * np.prod(np.arange(1, 2 * j, 2)) / (rootpi / np.sqrt(2.))
                           for j in range(_kde_lmax_bandwidth - 1, 1, -1)])


def _bandwidth_fixed_point(h, N, I, logI, a2):
    if h <= 0:
        return h - 1

    f = 2 * np.pi ** (2 * _kde_lmax_bandwidth) * np.dot(a2,
                                                        np.exp(_kde_lmax_bandwidth * logI - I * (pisquared * h ** 2)))
    for j, const in zip(range(_kde_lmax_bandwidth - 1, 1, -1), _kde_consts_1d):
        try:
            t_j = (const / N / f) ** (2 / (3. + 2 * j))
        except:
            print(f, h, N, j)
            raise
        f = 2 * np.pi ** (2 * j) * np.dot(a2, np.exp(j * logI - I * (pisquared * t_j)))
        if not f:
            raise Exception('zero f in _bandwidth_fixed_point (non-convergence)')
    return h - (2 * N * rootpi * f) ** (-1. / 5)


def bin_samples(samples, range_min=None, range_max=None, nbins=2046, edge_fac=0.1):
    mx = np.max(samples)
    mn = np.min(samples)
    delta = mx - mn
    if range_min is None:
        range_min = mn - delta * edge_fac
    if range_max is None:
        range_max = mx + delta * edge_fac
    R = range_max - range_min
    dx = R / (nbins - 1)
    bins = (samples - range_min) / dx
    return bins.astype(int), R


def gaussian_kde_bandwidth(samples, Neff=None, range_min=None, range_max=None, nbins=2046):
    if Neff is None:
        Neff = np.count_nonzero(np.diff(samples)) + 1
    bins, R = bin_samples(samples, range_min, range_max, nbins)
    data = np.bincount(bins, minlength=nbins)
    h = gaussian_kde_bandwidth_binned(data, Neff)
    if h is None:
        return None
    else:
        return h * R


def gaussian_kde_bandwidth_binned(data, Neff, a=None):
    """
     Return optimal standard kernel bandwidth assuming data is binned from Neff independent samples
     Return value is the bandwidth in units of the range of data (i.e. multiply by max(data)-min(data)),
     or None if failed

     Uses Improved Sheather-Jones algorithm from
     Kernel density estimation via diffusion: Z. I. Botev, J. F. Grotowski, and D. P. Kroese (2010)
     Annals of Statistics, Volume 38, Number 5, pages 2916-2957.
     http://arxiv.org/abs/1011.2602
    """
    I = np.arange(1, data.size) ** 2
    logI = np.log(I)
    if a is None:
        a = fftpack.dct(data / np.sum(data))
    a2 = (a[1:] / 2) ** 2
    try:
        n_scaling = Neff ** (-1. / 5)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            hfrac = 0.53 * n_scaling  #
            hfrac = fsolve(_bandwidth_fixed_point, hfrac, (Neff, I, logI, a2), xtol=hfrac / 20, factor=1)[0]
        if hfrac < 0.019 * n_scaling:
            # may be finding second solution, check with brent
            try:
                hfrac = brentq(_bandwidth_fixed_point, 0.019 * n_scaling, 0.5, (Neff, I, logI, a2), xtol=hfrac / 20)
            except Exception:
                # Could get sign error for the bounds in brentq, in which case small answer may be correct
                # or the method has failed (e.g. flat distribution between two bounds)
                pass
        return hfrac
    except Exception as e:
        logging.warning('1D auto bandwidth failed. Using fallback: %s' % e)
        return None


# 2D functions

K = np.array(
    [1 / np.sqrt(2 * np.pi)] + [(-1) ** j * np.prod(np.arange(1, 2 * j, 2)) / np.sqrt(2 * np.pi) for j in range(1, 5)])
Kodd = np.array([1] + [np.prod(np.arange(1, 2 * j, 2)) / 2. ** (j + 1) / np.sqrt(np.pi) for j in range(1, 9)])


class KernelOptimizer2D:
    def __init__(self, data, Neff, correlation, do_correlation=True, fallback_t=None):
        size = data.shape[0]
        if size != data.shape[1]:
            raise ValueError('KernelOptimizer2D only handles square arrays currently')
        self.a2 = dct2d(data / np.sum(data))[1:, 1:] ** 2
        self.I = np.arange(1, size, dtype=np.float64) ** 2
        self.logI = np.log(self.I)
        self.do_correlation = do_correlation
        if do_correlation:
            self.aFFT = np.fft.fft2(data[:, :] / np.sum(data))
            self.aFFT *= np.conj(self.aFFT)
        self.N = Neff
        self.corr = correlation
        try:
            # t is the bandwidth squared (used for estimating moments), calculated using fixed point
            self.t_star = brentq(self._bandwidth_fixed_point_2D, 0, 0.1, xtol=0.001 ** 2)
            # noinspection PyTypeChecker
            if fallback_t and self.t_star > 0.01 and self.t_star > 2 * fallback_t:
                # For 2D distributions with boundaries, fixed point can overestimate significantly
                logging.debug('KernelOptimizer2D Using fallback (t* > 2*t_gallback)')
                self.t_star = fallback_t
        except Exception:
            if fallback_t is not None:
                # Note the fallback result actually appears to be better in some cases,
                # e.g. Gaussian with four cuts
                logging.debug('2D kernel density optimizer using fallback plugin width %s' % (np.sqrt(fallback_t)))
                self.t_star = fallback_t
            else:
                raise

    def _bandwidth_fixed_point_2D(self, t):
        sum_func = self.func2d([0, 2], t) + self.func2d([2, 0], t) + 2 * self.func2d([1, 1], t)
        time = (2 * np.pi * self.N * sum_func) ** (-1. / 3)
        return (t - time) / time

    def psi(self, s, time):
        w = -self.I * (pisquared * time)
        wx = np.exp(w + self.logI * s[0])
        wy = np.exp(w + self.logI * s[1])
        return (-1) ** np.sum(s) * wy.dot(self.a2).dot(wx.T) * np.pi ** (2 * np.sum(s)) / 4

    def func2d(self, s, t):
        sums = np.sum(s)
        if sums <= 4:
            sum_func = self.func2d([s[0] + 1, s[1]], t) + self.func2d([s[0], s[1] + 1], t)
            const = (1 + 0.5 ** (sums + 1)) / 3
            time = (-2 * const * K[s[0]] * K[s[1]] / self.N / sum_func) ** (1. / (2 + sums))
            return self.psi(s, time)
        else:
            return self.psi(s, t)

    def func2d_odd(self, s, t):
        sums = np.sum(s)
        if sums <= 8:
            sum_func = self.func2d_odd([s[0] + 2, s[1]], t) + self.func2d_odd([s[0], s[1] + 2], t)
            const = 8 * (1 - 2. ** (-sums - 1)) / 3.
            # recall time is h^2
            time = (const * self.p00 * Kodd[s[0]] * Kodd[s[1]] / self.N ** 2 / sum_func ** 2) ** (1. / (3 + sums))
            return self.psi_odd(s, time)
        else:
            return self.psi_odd(s, t)

    def psi_odd(self, s, time):
        f = np.fft.fftfreq(self.aFFT.shape[0], d=1. / self.aFFT.shape[0])
        w = np.exp(-f ** 2 * (4 * pisquared * time))
        wx = w * f ** s[0]
        wy = w * f ** s[1]
        return wy.dot(self.aFFT).real.dot(wx.T) * (2 * np.pi) ** (np.sum(s))

    def AMISE(self, cov, corr=None):
        hx = cov[0]
        hy = cov[1]
        if corr is not None:
            c = corr
        else:
            c = cov[2]
        var = 1. / (4 * np.pi * hx * hy * np.sqrt(1 - c ** 2) * self.N)
        bias = 0.25 * (
                hx ** 4 * self.p[4, 0] + hy ** 4 * self.p[0, 4]
                + 2 * hx ** 2 * hy ** 2 * self.p[2, 2] * (2 * c ** 2 + 1)
                + 4 * c * hx * hy * (hx ** 2 * self.p[3, 1] + hy ** 2 * self.p[1, 3]))
        if bias < 0:
            raise Exception("bias not positive definite")
        return var + bias

    def get_h(self, do_correlation=None):
        if do_correlation is None:
            do_correlation = self.do_correlation
        p = np.zeros((5, 5))
        fixed = False
        tpsi = self.t_star
        if fixed:
            p_02 = self.psi([0, 2], tpsi)
            p_20 = self.psi([2, 0], tpsi)
            p_11 = self.psi([1, 1], tpsi)
        else:
            p_02 = self.func2d([0, 2], tpsi)
            p_20 = self.func2d([2, 0], tpsi)
            p_11 = self.func2d([1, 1], tpsi)
        h_x = (p_02 ** (3. / 4) / (4 * np.pi * self.N * p_20 ** (3. / 4) * (p_11 + np.sqrt(p_20 * p_02)))) ** (1. / 6)
        h_y = (p_20 ** (3. / 4) / (4 * np.pi * self.N * p_02 ** (3. / 4) * (p_11 + np.sqrt(p_20 * p_02)))) ** (1. / 6)
        corr = 0
        if not do_correlation:
            return h_x, h_y, corr

        p[0, 4] = p_02
        p[4, 0] = p_20
        p[2, 2] = p_11

        # p[0, 0] = self.psi([0, 0], tpsi)
        # self.p00 = p[0, 0]
        # p[1, 3] = self.psi_odd([1, 3], tpsi)
        # p[3, 1] = self.psi_odd([3, 1], tpsi)

        p[0, 0] = self.func2d([0, 0], tpsi)
        self.p00 = p[0, 0]
        p[1, 3] = self.func2d_odd([1, 3], tpsi)
        p[3, 1] = self.func2d_odd([3, 1], tpsi)

        self.p = p
        AMISE = self.AMISE(np.array([h_x, h_y, 0]))
        if self.corr:
            try:
                res = minimize(self.AMISE, np.array([h_x, h_y]) / np.sqrt(1 - abs(self.corr)), (self.corr,),
                               method='TNC', bounds=[(0.001, 0.3), (0.001, 0.3)])
                if res.success:
                    AMISEcorr = self.AMISE(res.x, self.corr)
                    if AMISEcorr < AMISE:
                        h_x, h_y = res.x
                        corr = self.corr
                        AMISE = AMISEcorr

            except:
                logging.debug('AMISE fixed correlation optimization failed')
                pass
        try:
            res = minimize(self.AMISE, np.array([h_x, h_y, self.corr]), (None,), method='TNC',
                           bounds=[(0.001, 0.3), (0.001, 0.3), (-0.99, 0.99)])
            if res.success:
                AMISEopt = self.AMISE(res.x)
                if AMISEopt < AMISE * 0.9:
                    h_x, h_y, corr = res.x
        except:
            logging.debug('AMISE optimization failed')
            pass
        return h_x, h_y, corr

    def get_hdiag(self):
        return self.get_h(do_correlation=False)



# This provides base classes for handling backwards compatibility with renamed or removed attributes

import re
import logging


def _convert_camel(name):
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()


def _map_name(obj, name):
    try:
        return object.__getattribute__(obj, name), name
    except AttributeError:
        pass
    _old = obj.__class__.__dict__.get('_deprecated')
    if _old and name in _old:
        newname = _old.get(name)
        if newname is None:
            return None, None
    else:
        newname = _convert_camel(name)
    try:
        return object.__getattribute__(obj, newname), newname
    except AttributeError:
        return None


class _BaseObject:
    # Compatibility of pep_8_style and camelCase for backwards compatibility

    _fail_on_not_exist = False

    def __getattribute__(self, name):
        if name.startswith('__'):
            return object.__getattribute__(self, name)
        res = _map_name(self, name)
        if res is None:
            raise AttributeError('%s is not a valid attribute for class %s' % (name, self.__class__))
        value, newname = res
        if newname is None:
            logging.warning("%s is removed and will be ignored" % name)
            return None
        if newname is not name:
            logging.warning("%s is deprecated, use %s" % (name, newname))
        return value

    def __setattr__(self, name, value):

        res = _map_name(self, name)
        if res is None:
            if object.__getattribute__(self, '_fail_on_not_exist'):
                raise AttributeError('Unknown attribute %s for class %s' % (name, self.__class__))
            newname = name
        else:
            _, newname = res
        if newname is None:
            logging.warning("%s is removed and will be ignored" % name)
            return
        object.__setattr__(self, newname, value)



from getdist import plots
from matplotlib import cm


# Simple style that uses matplotlib's default color table for contours and lines

class DefaultColorsPlotter(plots.GetDistPlotter):
    # noinspection PyUnresolvedReferences
    def set_default_settings(self):
        s = plots.GetDistPlotSettings()
        s.solid_colors = cm.tab10
        s.line_styles = cm.tab10
        s.colormap_scatter = 'viridis'
        self.settings = s


style_name = 'tab10'
plots.add_plotter_style(style_name, DefaultColorsPlotter)



import numpy as np
from getdist.densities import Density1D, Density2D
from getdist.paramnames import ParamNames
from getdist.mcsamples import MCSamples
import copy


def make_2D_Cov(sigmax, sigmay, corr):
    return np.array([[sigmax ** 2, sigmax * sigmay * corr], [sigmax * sigmay * corr, sigmay ** 2]])


class MixtureND:
    """
    Gaussian mixture model with optional boundary ranges. Includes functions for generating samples and projecting.
    """

    def __init__(self, means, covs, weights=None, lims=None, names=None, label='', labels=None):
        """
        :param means: list of y for each Gaussian in the mixture
        :param covs: list of covariances for the Gaussians in the mixture
        :param weights: optional weight for each component (defaults to equal weight)
        :param lims: optional list of hard limits for each parameter, [[x1min,x1max], [x2min,x2max]];
                     use None for no limit
        :param names: list of names (strings) for each parameter. If not set, set to "param1", "param2"...
        :param label: name for labelling this mixture
        :param labels: list of latex labels for each parameter. If not set, defaults to p_{1}, p_{2}...
        """

        self.means = np.asarray(means)
        self.dim = self.means.shape[1]
        self.covs = [np.array(cov) for cov in covs]
        self.invcovs = [np.linalg.inv(cov) for cov in self.covs]
        if weights is None:
            weights = [1. / len(means)] * len(means)
        self.weights = np.array(weights, dtype=np.float64)
        if np.sum(self.weights) <= 0:
            raise ValueError('Weight <= 0 in MixtureND')
        self.weights /= np.sum(weights)
        self.norms = (2 * np.pi) ** (0.5 * self.dim) * np.array([np.sqrt(np.linalg.det(cov)) for cov in self.covs])
        self.lims = lims
        self.paramNames = ParamNames(names=names, default=self.dim, labels=labels)
        self.names = self.paramNames.list()
        self.label = label
        self.total_mean = np.atleast_1d(np.dot(self.weights, self.means))
        self.total_cov = np.zeros((self.dim, self.dim))
        for mean, cov, weight, totmean in zip(self.means, self.covs, self.weights, self.total_mean):
            self.total_cov += weight * (cov + np.outer(mean - totmean, mean - totmean))

    def sim(self, size, random_state=None):
        """
        Generate an array of independent samples

        :param size: number of samples
        :param random_state: random number Generator or seed
        :return: 2D array of sample values
        """
        tot = 0
        res = []
        block = None
        random_state = np.random.default_rng(random_state)
        while True:
            for num, mean, cov in zip(random_state.multinomial(block or size, self.weights), self.means, self.covs):
                if num > 0:
                    v = random_state.multivariate_normal(mean, cov, size=num)
                    if self.lims is not None:
                        for i, (mn, mx) in enumerate(self.lims):
                            if mn is not None:
                                v = v[v[:, i] >= mn]
                            if mx is not None:
                                v = v[v[:, i] <= mx]
                    tot += v.shape[0]
                    res.append(v)
            if tot >= size:
                break
            if block is None:
                block = min(max(size, 100000), int(1.1 * (size * (size - tot))) // max(tot, 1) + 1)
        samples = np.vstack(res)
        if len(res) > 1:
            samples = random_state.permutation(samples)
        if tot != size:
            samples = samples[:-(tot - size), :]
        return samples

    def MCSamples(self, size, names=None, logLikes=False, random_state=None, **kwargs):
        """
        Gets a set of independent samples from the mixture as a  :class:`.mcsamples.MCSamples` object
        ready for plotting etc.

        :param size: number of samples
        :param names: set to override existing names
        :param logLikes: if True set the sample likelihood values from the pdf, if false, don't store log likelihoods
        :param random_state: random seed or Generator
        :return: a new :class:`.mcsamples.MCSamples` instance
        """
        samples = self.sim(size, random_state=random_state)
        if logLikes:
            loglikes = -np.log(self.pdf(samples))
        else:
            loglikes = None
        return MCSamples(samples=samples, loglikes=loglikes, paramNamesFile=copy.deepcopy(self.paramNames),
                         names=names, ranges=self.lims, **kwargs)

    def autoRanges(self, sigma_max=4, lims=None):
        res = []
        if lims is None:
            lims = self.lims
        if lims is None:
            lims = [(None, None) for _ in range(self.dim)]
        for i, (mn, mx) in enumerate(lims):
            covmin = None
            covmax = None
            if mn is None or mx is None:
                for mean, cov in zip(self.means, self.covs):
                    sigma = np.sqrt(cov[i, i])
                    xmin, xmax = mean[i] - sigma_max * sigma, mean[i] + sigma_max * sigma
                    if mn is not None:
                        xmax = max(xmax, mn + sigma_max * sigma)
                    if mx is not None:
                        xmin = min(xmin, mx - sigma_max * sigma)
                    covmin = min(xmin, covmin) if covmin is not None else xmin
                    covmax = max(xmax, covmax) if covmax is not None else xmax
            res.append((covmin if mn is None else mn, covmax if mx is None else mx))
        return res

    def pdf(self, x):
        """
        Calculate the PDF. Note this assumes x is within the boundaries (does not return zero outside)
        Result is also only normalized if no boundaries.

        :param x: array of parameter values to evaluate at
        :return: pdf at x
        """
        tot = None
        x = np.asarray(x)
        for i, (mean, icov, weight, norm) in enumerate(zip(self.means, self.invcovs, self.weights, self.norms)):
            dx = x - mean
            if len(x.shape) == 1:
                res = np.exp(-icov.dot(dx).dot(dx) / 2) / norm
            else:
                res = np.exp(-np.einsum('ik,km,im->i', dx, icov, dx) / 2) / norm
            if not i:
                tot = res * weight
            else:
                tot += res * weight
        return tot

    def pdf_marged(self, index, x, no_limit_marge=False):
        """
        Calculate the 1D marginalized PDF. Only works if no other parameter limits are marginalized

        :param index: index or name of parameter
        :param x: value to evaluate PDF at
        :param no_limit_marge: if true don't raise an error if mixture has limits
        :return: marginalized 1D pdf at x
        """
        if isinstance(index, str):
            index = self.names.index(index)
        if not no_limit_marge:
            self.checkNoLimits([index])
        tot = None
        for i, (mean, cov, weight) in enumerate(zip(self.means, self.covs, self.weights)):
            dx = x - mean[index]
            var = cov[index, index]
            res = np.exp(-dx ** 2 / var / 2) / np.sqrt(2 * np.pi * var)
            if not i:
                tot = res * weight
            else:
                tot += res * weight
        return tot

    def density1D(self, index=0, num_points=1024, sigma_max=4, no_limit_marge=False):
        """
        Get 1D marginalized density. Only works if no hard limits in other parameters.

        :param index: parameter name or index
        :param num_points: number of grid points to evaluate PDF
        :param sigma_max: maximum number of standard deviations away from y to include in computed range
        :param no_limit_marge: if true don't raise error if limits on other parameters
        :return: :class:`~.densities.Density1D` instance
        """
        if isinstance(index, str):
            index = self.names.index(index)
        if not no_limit_marge:
            self.checkNoLimits([index])
        mn, mx = self.autoRanges(sigma_max)[index]
        x = np.linspace(mn, mx, num_points)
        like = self.pdf_marged(index, x)
        return Density1D(x, like)

    def density2D(self, params=None, num_points=1024, xmin=None, xmax=None, ymin=None, ymax=None, sigma_max=5):
        """
        Get 2D marginalized density for a pair of parameters.

        :param params: list of two parameter names or indices to use. If already 2D, can be None.
        :param num_points: number of grid points for evaluation
        :param xmin: optional lower value for first parameter
        :param xmax: optional upper value for first parameter
        :param ymin: optional lower value for second parameter
        :param ymax: optional upper value for second parameter
        :param sigma_max: maximum number of standard deviations away from mean to include in calculated range
        :return: :class:`~.densities.Density2D` instance
        """
        if self.dim > 2 or params is not None or not isinstance(self, Mixture2D):
            mixture = self.marginalizedMixture(params=params)
        elif self.dim != 2:
            raise Exception('density2D requires at least two dimensions')
        else:
            mixture = self

        # noinspection PyProtectedMember
        return mixture._density2D(num_points=num_points, xmin=xmin, xmax=xmax, ymin=ymin,
                                  ymax=ymax, sigma_max=sigma_max)

    def _params_to_indices(self, params):
        indices = []
        if params is None:
            params = self.names
        for p in params:
            if isinstance(p, str):
                indices.append(self.names.index(p))
            elif hasattr(p, 'name'):
                indices.append(self.names.index(p.name))
            else:
                indices.append(p)
        return indices

    def marginalizedMixture(self, params, label=None, no_limit_marge=False) -> 'MixtureND':
        """
        Calculates a reduced mixture model by marginalization over unwanted parameters

        :param params: array of parameter names or indices to retain.
                       If none, will simply return a copy of this mixture.
        :param label: optional label for the marginalized mixture
        :param no_limit_marge: if true don't raise an error if mixture has limits.
        :return: a new marginalized  :class:`MixtureND` instance
        """

        indices = self._params_to_indices(params)
        if not no_limit_marge:
            self.checkNoLimits(indices)
        indices = np.array(indices)
        if self.names is not None:
            names = [self.names[i] for i in indices]
        else:
            names = None
        if self.lims is not None:
            lims = [self.lims[i] for i in indices]
        else:
            lims = None
        if label is None:
            label = self.label
        covs = [cov[np.ix_(indices, indices)] for cov in self.covs]
        means = [mean[indices] for mean in self.means]
        if len(indices) == 2:
            tp = Mixture2D
        else:
            tp = MixtureND
        mixture = tp(means, covs, self.weights, lims=lims,
                     names=names, label=label)
        mixture.paramNames.setLabelsAndDerivedFromParamNames(self.paramNames)
        return mixture

    def conditionalMixture(self, fixed_params, fixed_param_values, label=None):
        """
        Returns a reduced conditional mixture model for the distribution when certainly parameters are fixed.

        :param fixed_params: list of names or numbers of parameters to fix
        :param fixed_param_values:  list of values for the fixed parameters
        :param label: optional label for the new mixture
        :return: A new :class:`MixtureND` instance with cov_i = Projection(Cov_i^{-1})^{-1} and shifted conditional y
        """

        fixed_params = self._params_to_indices(fixed_params)
        self.checkNoLimits(fixed_params)
        keep_params = [i for i in range(self.dim) if i not in fixed_params]
        if not len(keep_params):
            raise ValueError('conditionalMixture must leave at least one non-fixed parameter')
        new_means = []
        new_covs = []
        new_weights = []
        for mean, cov, invcov, weight in zip(self.means, self.covs, self.invcovs, self.weights):
            deltas = np.asarray(fixed_param_values) - mean[fixed_params]
            new_cov = np.linalg.inv(invcov[np.ix_(keep_params, keep_params)])
            new_mean = mean[keep_params] - new_cov.dot(invcov[np.ix_(keep_params, fixed_params)].dot(deltas))
            if len(self.weights) == 1 and False:
                logw = 0
            else:
                logw = invcov[np.ix_(fixed_params, fixed_params)].dot(deltas).dot(deltas) \
                       + np.log(np.linalg.det(cov[np.ix_(fixed_params, fixed_params)]
                                              - cov[np.ix_(fixed_params, keep_params)]
                                              .dot(np.linalg.inv(cov[np.ix_(keep_params, keep_params)])
                                                   .dot(cov[np.ix_(keep_params, fixed_params)]))))
            new_weights.append(logw)
            new_means.append(new_mean)
            new_covs.append(new_cov)

        new_weights = np.exp(-(np.asarray(new_weights) - min(new_weights)) / 2)
        if self.names is not None:
            names = [self.names[i] for i in keep_params]
        else:
            names = None
        mixture = MixtureND(new_means, new_covs, new_weights, names=names, label=label)
        mixture.paramNames.setLabelsAndDerivedFromParamNames(self.paramNames)
        return mixture

    def checkNoLimits(self, keep_params):
        if self.lims is None:
            return
        for i, lim in enumerate(self.lims):
            if i not in keep_params and (lim[0] is not None or lim[1] is not None):
                raise Exception(
                    'In general can only marginalize analytically if no hard boundary limits: ' + self.label)

    def getUpper(self, name):
        if self.lims is None:
            return None
        return self.lims[self.names.index(name)][1]

    def getLower(self, name):
        if self.lims is None:
            return None
        return self.lims[self.names.index(name)][1]


class Mixture2D(MixtureND):
    """
    Gaussian mixture model in 2D with optional boundaries for fixed x and y ranges
    """

    def __init__(self, means, covs, weights=None, lims=None, names=('x', 'y'),
                 xmin=None, xmax=None, ymin=None, ymax=None, **kwargs):
        """
        :param means: list of y for each Gaussian in the mixture
        :param covs: list of covariances for the Gaussians in the mixture. Instead of 2x2 arrays,
                     each cov can also be a list of [sigma_x, sigma_y, correlation] parameters
        :param weights: optional weight for each component (defaults to equal weight)
        :param lims: optional list of hard limits for each parameter, [[x1min,x1max], [x2min,x2max]];
                     use None for no limit
        :param names: list of names (strings) for each parameter. If not set, set to x, y
        :param xmin: optional lower hard bound for x
        :param xmax: optional upper hard bound for x
        :param ymin: optional lower hard bound for y
        :param ymax: optional upper hard bound for y
        :param kwargs: arguments passed to :class:`MixtureND`
        """

        if lims is not None:
            limits = self._updateLimits(lims, xmin, xmax, ymin, ymax)
        else:
            limits = [(xmin, xmax), (ymin, ymax)]
        mats = []
        for cov in covs:
            if isinstance(cov, (list, tuple)) and len(cov) == 3 and not isinstance(cov[0], (list, tuple)):
                mats.append(make_2D_Cov(*cov))
            else:
                mats.append(cov)
        MixtureND.__init__(self, means, mats, weights, limits, names=names, **kwargs)

    def _updateLimits(self, lims, xmin=None, xmax=None, ymin=None, ymax=None):
        xmin = xmin if xmin is not None else lims[0][0]
        xmax = xmax if xmax is not None else lims[0][1]
        ymin = ymin if ymin is not None else lims[1][0]
        ymax = ymax if ymax is not None else lims[1][1]
        return [(xmin, xmax), (ymin, ymax)]

    def _density2D(self, num_points=1024, xmin=None, xmax=None, ymin=None, ymax=None, sigma_max=5):
        lims = self._updateLimits(self.lims, xmin, xmax, ymin, ymax)
        (xmin, xmax), (ymin, ymax) = self.autoRanges(sigma_max, lims=lims)
        x = np.linspace(xmin, xmax, num_points)
        y = np.linspace(ymin, ymax, num_points)
        xx, yy = np.meshgrid(x, y)
        like = self.pdf(xx, yy)
        return Density2D(x, y, like)

    def pdf(self, x, y=None):
        """
        Calculate the PDF. Note this assumes x and y are within the boundaries (does not return zero outside)
        Result is also only normalized if no boundaries

        :param x: value of x to evaluate pdf
        :param y: optional value of y to evaluate pdf. If not specified, returns 1D marginalized value for x.
        :return: value of pdf at x or x,y
        """
        if y is None:
            return super().pdf(x)
        tot = None
        for i, (mean, icov, weight, norm) in enumerate(zip(self.means, self.invcovs, self.weights, self.norms)):
            dx = x - mean[0]
            dy = y - mean[1]
            res = np.exp(-(dx ** 2 * icov[0, 0] + 2 * dx * dy * icov[0, 1] + dy ** 2 * icov[1, 1]) / 2) / norm
            if not i:
                tot = res * weight
            else:
                tot += res * weight
        return tot


class Gaussian2D(Mixture2D):
    """
    Simple special case of a 2D Gaussian mixture model with only one Gaussian component
    """

    def __init__(self, mean, cov, **kwargs):
        """
        :param mean: 2 element array with mean
        :param cov: 2x2 array of covariance, or list of [sigma_x, sigma_y, correlation] values
        :param kwargs: arguments passed to :class:`Mixture2D`
        """
        super().__init__([mean], [cov], **kwargs)


class GaussianND(MixtureND):
    """
        Simple special case of a Gaussian mixture model with only one Gaussian component
    """

    def __init__(self, mean, cov, is_inv_cov=False, **kwargs):
        """
        :param mean: array specifying y of parameters
        :param cov: covariance matrix (or filename of text file with covariance matrix)
        :param is_inv_cov: set True if cov is actually an inverse covariance
        :param kwargs: arguments passed to :class:`MixtureND`
        """
        if isinstance(mean, str):
            mean = np.loadtxt(mean)
        if isinstance(cov, str):
            cov = np.loadtxt(cov)
        if is_inv_cov:
            cov = np.linalg.inv(cov)
        super().__init__([mean], [cov], **kwargs)


class Mixture1D(MixtureND):
    """
    Gaussian mixture model in 1D with optional boundaries for fixed ranges
    """

    def __init__(self, means, sigmas, weights=None, lims=None, name='x',
                 xmin=None, xmax=None, **kwargs):
        """
        :param means: array of y for each component
        :param sigmas: array of standard deviations for each component
        :param weights: weights for each component (defaults to equal weight)
        :param lims: optional array limits for each component
        :param name: parameter name (default 'x')
        :param xmin: optional lower limit
        :param xmax:  optional upper limit
        :param kwargs: arguments passed to :class:`MixtureND`
        """
        if lims is not None:
            limits = [(xmin if xmin is not None else lims[0], xmax if xmax is not None else lims[1])]
        else:
            limits = [(xmin, xmax)]
        covs = [np.atleast_2d(sigma ** 2) for sigma in sigmas]
        means = [[mean] for mean in means]
        MixtureND.__init__(self, means, covs, weights, limits, names=[name], **kwargs)

    def pdf(self, x):
        return self.pdf_marged(0, x)


class Gaussian1D(Mixture1D):
    """
    Simple 1D Gaussian
    """

    def __init__(self, mean, sigma, **kwargs):
        """
        :param mean: mean
        :param sigma:  standard deviation
        :param kwargs:  arguments passed to :class:`Mixture1D`
        """
        super().__init__([mean], [sigma], **kwargs)


class RandomTestMixtureND(MixtureND):
    """
    class for randomly generating an N-D gaussian mixture for testing (a mixture with random parameters, not random
    samples from the mixture).
    """

    def __init__(self, ndim=4, ncomponent=1, names=None, weights=None, seed=None, label='RandomMixture'):
        """
        :param ndim: number of dimensions
        :param ncomponent: number of components
        :param names: names for the parameters
        :param weights: weights for each component
        :param seed:  random seed or Generator
        :param label: label for the generated mixture
        """
        random_state = np.random.default_rng(seed)
        covs = []
        for _ in range(ncomponent):
            A = random_state.random((ndim, ndim))
            covs.append(np.dot(A, A.T))
        super().__init__(random_state.random((ncomponent, ndim)), covs, weights=weights,
                         lims=None, names=names, label=label)


def randomTestMCSamples(ndim=4, ncomponent=1, nsamp=10009, nMCSamples=1, seed=10, names=None, labels=None):
    """
    get a list of MCSamples instances with random samples from random covariances and y
    """
    if names is None:
        names = ["x%s" % i for i in range(ndim)]
    if labels is None:
        labels = ["x_{%s}" % i for i in range(ndim)]
    seed = np.random.default_rng(seed)
    return [RandomTestMixtureND(ndim, ncomponent, names, seed=seed).MCSamples(
        nsamp, labels=labels, name_tag='Sim %s' % (i + 1), random_state=seed) for i in range(nMCSamples)]



# JT 2017-19

from importlib import import_module
from copy import deepcopy
import logging
from numbers import Number
import numpy as np
import os
from typing import Mapping, Sequence

# Conventions
_label = "label"
_prior = "prior"
_theory = "theory"
_params = "params"
_likelihood = "likelihood"
_sampler = "sampler"
_p_label = "latex"
_p_dist = "dist"
_p_value = "value"
_p_derived = "derived"
_p_renames = "renames"
_separator = "__"
_minuslogprior = "minuslogprior"
_prior_1d_name = "0"
_chi2 = "chi2"
_weight = "weight"
_minuslogpost = "minuslogpost"
_post = "post"


def cobaya_params_file(root):
    file = root + ('' if root.endswith((os.sep, "/")) else '.') + 'updated.yaml'
    if os.path.exists(file):
        return file
    else:
        file = root + ('' if root.endswith((os.sep, "/")) else '__') + 'full.yaml'
        if os.path.exists(file):
            return file
    return None


def yaml_file_or_dict(file_or_dict) -> Mapping:
    if isinstance(file_or_dict, str):
        from getdist.yaml_tools import yaml_load_file
        return yaml_load_file(file_or_dict)
    elif isinstance(file_or_dict, Mapping):
        return file_or_dict
    else:
        raise ValueError('Cobya parameter input must be a dictionary or filename')


def MCSamplesFromCobaya(info, collections, name_tag=None,
                        ignore_rows=0, ini=None, settings=None):
    """
    Creates a set of samples from Cobaya's output.
    Parameter names, ranges and labels are taken from the "info" dictionary
    (always use the "updated" one generated by `cobaya.run`).

    For a description of the various analysis settings and default values see
    `analysis_defaults.ini <https://getdist.readthedocs.org/en/latest/analysis_settings.html>`_.

    :param collections: collection(s) of samples from Cobaya
    :param info: info dictionary, common to all collections
                 (use the "updated" one, returned by `cobaya.run`)
    :param name_tag: name for this sample to be shown in the plots' legend
    :param ignore_rows: initial samples to skip, number (`int>=1`) or fraction (`float<1`)
    :param ini: The name of a .ini file with analysis settings to use
    :param settings: dictionary of analysis settings to override defaults
    :return: The :class:`MCSamples` instance
    """

    if hasattr(collections, "data"):
        collections = [collections]
    # Check consistency between collections
    try:
        columns = list(collections[0].data)
    except AttributeError:
        raise TypeError(
            "The second argument does not appear to be a (list of) samples `Collection`.")
    if not all(list(c.data) == columns for c in collections[1:]):
        raise ValueError("The given collections don't have the same columns.")
    # Check consistency with info
    info_params = get_info_params(info)
    # if skip burn in *has already been done*
    skip = info.get(_post, {}).get("skip", 0)
    if ignore_rows != 0 and skip != 0:
        logging.warning("You are asking for rows to be ignored (%r), but some (%r) were "
                        "already ignored in the original chain.", ignore_rows, skip)
    var_params = [k for k, v in info_params.items() if
                  is_sampled_param(v) or is_derived_param(v)]
    assert set(columns[2:]) == set(var_params), (
            "Info and collection(s) are not compatible, because their parameters differ: "
            "the collection(s) have %r and the info has %r. " % (
                columns[2:], var_params) +
            "Are you sure that you are using an *updated* info dictionary "
            "(i.e. the output of `cobaya.run`)?")
    # We need to use *collection* sorting, not info sorting!
    names = [p + ("*" if is_derived_param(info_params[p]) else "")
             for p in columns[2:]]
    labels = [(info_params[p] or {}).get(_p_label, p) for p in columns[2:]]
    ranges = {p: get_range(info_params[p]) for p in
              info_params}  # include fixed parameters not in columns
    renames = {p: info_params.get(p, {}).get(_p_renames, []) for p in columns[2:]}
    samples = [c[c.data.columns[2:]].values for c in collections]
    weights = [c[_weight].values for c in collections]
    loglikes = [c[_minuslogpost].values for c in collections]
    sampler = get_sampler_type(info)
    label = get_sample_label(info)
    from getdist.mcsamples import MCSamples
    return MCSamples(samples=samples, weights=weights, loglikes=loglikes, sampler=sampler,
                     names=names, labels=labels, ranges=ranges, renames=renames,
                     ignore_rows=ignore_rows, name_tag=name_tag, label=label, ini=ini,
                     settings=settings)


def str_to_list(x):
    return [x] if isinstance(x, str) else x


def get_info_params(info):
    """
    Extracts parameter info from the new yaml format.
    """
    info = yaml_file_or_dict(info)
    # Prune fixed parameters
    info_params = info.get(_params)
    info_params_full = dict()
    for p, pinfo in info_params.items():
        info_params_full[p] = info_params[p]
    # Add prior and likelihoods
    priors = [_prior_1d_name] + list(info.get(_prior) or [])
    likes = list(info.get(_likelihood))
    # Account for post
    remove = info.get(_post, {}).get("remove", {})
    for param in remove.get(_params, []) or []:
        info_params_full.pop(param, None)
    for like in str_to_list(remove.get(_likelihood) or []):
        likes.remove(like)
    for prior in str_to_list(remove.get(_prior)) or []:
        priors.remove(prior)
    add = info.get(_post, {}).get("add", {})
    # Adding derived params and updating 1d priors
    for param, pinfo in add.get(_params, {}).items():
        pinfo_old = info_params_full.get(param, {})
        pinfo_old.update(pinfo)
        info_params_full[param] = pinfo_old
    likes += list(add.get(_likelihood, []))
    priors += list(add.get(_prior, []))
    # Add the prior and the likelihood as derived parameters
    info_params_full[_minuslogprior] = {_p_label: r"-\log\pi"}
    for prior in priors:
        info_params_full[_minuslogprior + _separator + prior] = {
            _p_label: r"-\log\pi_\mathrm{" + prior.replace("_", r"\ ") + r"}"}
    info_params_full[_chi2] = {_p_label: r"\chi^2"}
    for like in likes:
        info_params_full[_chi2 + _separator + like] = {
            _p_label: r"\chi^2_\mathrm{" + like.replace("_", r"\ ") + r"}"}
    return info_params_full


# noinspection PyUnboundLocalVariable
def get_range(param_info):
    # Sampled
    if is_sampled_param(param_info):
        if isinstance(param_info[_prior], Sequence) and len(param_info[_prior]) == 2:
            param_info[_prior] = \
                {lim: n for lim, n in zip(["min", "max"], param_info[_prior])}
        elif not isinstance(param_info[_prior], Mapping):
            raise ValueError(
                "Format of prior not recognised: %r. " % param_info[_prior] +
                "Use '[min, max]' or a dictionary following Cobaya's documentation.")
        info_lims = dict((tag, param_info[_prior].get(tag))
                         for tag in ["min", "max", "loc", "scale"])
        if info_lims["min"] is not None or info_lims["max"] is not None:
            lims = [param_info[_prior].get("min"), param_info[_prior].get("max")]
        elif info_lims["loc"] is not None or info_lims["scale"] is not None:
            dist = param_info[_prior].pop(_p_dist, "uniform")
            pdf_dist = getattr(import_module("scipy.stats", dist), dist)
            lims = pdf_dist.interval(1, **param_info[_prior])
    # Derived
    elif is_derived_param(param_info):
        lims = (lambda i: [i.get("min", -np.inf), i.get("max", np.inf)])(param_info or {})
    # Fixed
    else:
        value = fixed_value(param_info)
        try:
            value = float(value)
        except ValueError:
            # e.g. lambda function values
            lims = (lambda i: [i.get("min", -np.inf), i.get("max", np.inf)])(
                param_info or {})
        else:
            lims = (value, value)
    return lims[0] if lims[0] != -np.inf else None, lims[1] if lims[1] != np.inf else None


def fixed_value(info_param):
    """
    Returns True if the parameter has been fixed to a value or through a function.
    """
    return expand_info_param(info_param).get(_p_value, None)


def is_fixed_param(info_param):
    """
    Returns True if the parameter has been fixed to a value or through a function.
    """
    return fixed_value(info_param) is not None


def is_parameter_with_range(info_param):
    value = fixed_value(info_param)
    return value is None or isinstance(value, Number) or is_derived_param(info_param)


def is_sampled_param(info_param):
    """
    Returns True if the parameter has a prior.
    """
    return _prior in expand_info_param(info_param)


def is_derived_param(info_param):
    """
    Returns True if the parameter is saved as a derived one.
    """
    return expand_info_param(info_param).get(_p_derived, False)


def expand_info_param(info_param):
    """
    Expands the info of a parameter, from the user friendly, shorter format
    to a more unambiguous one.
    """
    if not isinstance(info_param, Mapping):
        if info_param is None:
            info_param = {}
        else:
            info_param = {_p_value: info_param}
    else:
        info_param = deepcopy(info_param)
    if all((f not in info_param) for f in [_prior, _p_value, _p_derived]):
        info_param[_p_derived] = True
    # Dynamical input parameters: save as derived by default
    value = info_param.get(_p_value, None)
    if isinstance(value, str) or callable(value):
        info_param[_p_derived] = info_param.get(_p_derived, True)
    return info_param


def get_sampler_type(filename_or_info, default_sampler_for_chain_type="mcmc"):
    sampler = list(yaml_file_or_dict(filename_or_info).get(_sampler, [
        default_sampler_for_chain_type]))[0]
    return {"mcmc": "mcmc", "polychord": "nested", "minimize": "minimize"}[sampler]


def get_sample_label(filename_or_info):
    return yaml_file_or_dict(filename_or_info).get(_label, None)


def get_burn_removed(filename_or_info):
    info = get_info_params(filename_or_info)
    # if skip burn in *has already been done*
    return info.get(_post, {}).get("skip", 0)



import os

try:
    from getdist.plots import get_subplot_plotter
except ImportError:
    import sys

    sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), '..', '..')))
    from getdist.plots import get_subplot_plotter
import matplotlib.pyplot as plt
import numpy as np
from getdist.gaussian_mixtures import Mixture2D, Mixture1D, Gaussian1D, Gaussian2D, make_2D_Cov
from getdist import chains
from getdist.mcsamples import BandwidthError
import logging

default_nsamp = 10000


def simFiles(prob, file_root, sample_lengths=(1000, 2000, 5000, 10000, 20000, 50000, 100000), text=True):
    for nn in sample_lengths:
        samples = prob.MCSamples(nn, logLikes=True)
        if text:
            samples.saveAsText(file_root + '_' + str(nn))
        else:
            samples.savePickle(file_root + '.py_mcsamples')


def compareSimPlot2D(g, samples, density, pars=('x', 'y')):
    g.plot_2d(samples, pars)
    density.normalize('max')
    levels = density.getContourLevels(contours=[0.68, 0.95])
    g.add_2d_density_contours(density, filled=False, color='C0', contour_levels=levels)
    levels = density.getContourLevels(contours=[0.2, 0.4, 0.6, 0.8])
    g.add_2d_density_contours(density, filled=False, color='C3', alpha=0.5, contour_levels=levels)


def compareSimPlot(g, samples, density, par='x', normalized=True):
    g.plot_1d(samples, par, colors=['C0'], normalized=normalized)
    density.normalize('integral' if normalized else 'max')
    plt.plot(density.x, density.P, color='C3')
    if normalized:
        plt.ylim(0, plt.gca().get_ylim()[1] * 1.1)


def plot1DSim(g, prob, nsamp=default_nsamp, settings=None):
    samps = prob.MCSamples(nsamp, settings=settings)
    compareSimPlot(g, samps, prob.density1D())


def plot2DSim(g, prob, nsamp=default_nsamp, settings=None):
    samps = prob.MCSamples(nsamp, settings=settings)
    compareSimPlot2D(g, samps, prob.density2D())


def compare1D(g, probs, nsamp=default_nsamp, settings=None):
    samples = []
    for i, prob in enumerate(probs):
        samps = prob.MCSamples(nsamp, settings=settings)
        samples.append(samps)

    g.make_figure(len(probs))
    for i, (samps, prob) in enumerate(zip(samples, probs)):
        g._subplot_number(i)
        compareSimPlot(g, samps, prob.density1D())
        g.add_text_left(prob.label, y=0.98, fontsize=8, verticalalignment='top')
    plt.subplots_adjust()


def compare2D(g, probs, nsamp=default_nsamp, settings=None):
    samples = []
    for i, prob in enumerate(probs):
        samps = prob.MCSamples(nsamp, settings=settings)
        samples.append(samps)

    g.make_figure(len(probs))
    for i, (samps, prob) in enumerate(zip(samples, probs)):
        g._subplot_number(i)
        compareSimPlot2D(g, samps, prob.density2D())
        g.add_text_left(prob.label, y=0.96, fontsize=8, verticalalignment='top')
    plt.subplots_adjust()


# noinspection PyUnboundLocalVariable
def get2DMises(prob, nsamp=default_nsamp, nsim=20, scales=np.arange(0.6, 1.5, 0.1), settings=None):
    # Get 2D MISE (mean integrated square error) as function of bandwidth scaling
    Mises = np.zeros(np.asarray(scales).size)
    for _ in range(nsim):
        samps = prob.MCSamples(nsamp, settings=settings)
        for i, scale in enumerate(scales):
            density = samps.get2DDensity('x', 'y', smooth_scale_2D=-scale)
            density.normalize()
            if i == 0:
                xgrid, ygrid = np.meshgrid(density.x, density.y)
                mean = prob.pdf(xgrid, ygrid)
                mean /= density.integrate(mean)
            Mises[i] += np.sum((mean - density.P) ** 2) / np.sum(mean ** 2)
    Mises /= (nsim - 1)
    return scales, Mises


def get1DMises(prob, nsamp=default_nsamp, nsim=50, scales=(0.6, 1.5, 0.1), settings=None):
    # Get 1D MISE (mean integrated square error) as function of bandwidth scaling
    Mises = np.zeros(np.asarray(scales).size)
    failures = 0
    for _ in range(nsim):
        samps = prob.MCSamples(nsamp, settings=settings)
        samps.raise_on_bandwidth_errors = False
        try:
            for i, scale in enumerate(scales):
                density = samps.get1DDensity('x', smooth_scale_1D=-scale)
                density.normalize()
                if i == 0:
                    mean = prob.pdf(density.x)
                    if prob.lims is not None:
                        mean /= density.integrate(mean)
                # noinspection PyUnboundLocalVariable
                Mises[i] += np.sum((mean - density.P) ** 2) / np.sum(mean ** 2)
        except BandwidthError:
            failures += 1
    Mises /= (nsim - failures - 1)
    if failures:
        logging.warning('%s failures for 1D bandwidth with %s' % (failures, prob.label))
    return scales, Mises


class Test1DDistributions:
    def __init__(self):
        self.gauss = Gaussian1D(0, 0.5, label='Gaussian')
        self.skew = Mixture1D([0, 1], [1, 0.4], [0.6, 0.4], label='skew')
        self.tailed = Mixture1D([0, 0], [1, 3], [0.8, 0.2], label='tailed')
        self.flat = Gaussian1D(0, 3, xmin=-1, xmax=2, label='flat')

        self.broad = Mixture1D([0, 0.3], [1, 2], [0.6, 0.4], label='broad')
        self.flat_top = Mixture1D([0, 1.5, 3], [1, 1, 1], [0.4, 0.2, 0.4], label='flat top')
        self.bimodal = []
        self.bimodal.append(Mixture1D([0, 2], [0.5, 0.5], [0.6, 0.4], label='bimodal 1'))
        self.bimodal.append(Mixture1D([0, 2], [0.2, 0.5], [0.5, 0.5], label='bimodal 2'))
        self.trimodal = []
        self.trimodal.append(Mixture1D([0, 2, 5], [0.2, 0.7, 0.4], label='trimodal'))
        self.cut_gaussians = self.cutGaussians()
        self.shape_set = [self.gauss, self.skew, self.tailed, self.broad, self.flat, self.flat_top]
        self.all = self.shape_set + self.bimodal + self.trimodal + self.cut_gaussians

    def cutGaussians(self, sigma=1, cut_x=(-1.5, -1, -0.5, 0, 1, 1.5)):
        return [Gaussian1D(0, sigma, xmin=cut, label=r'Gaussian [$x>%s$]' % cut) for cut in cut_x]

    def distributions(self):
        return self.all


class Test2DDistributions:
    def __init__(self):
        self.gauss = Gaussian2D([0, 0], (0.7, 1, 0.3), label='Gaussian')

        self.bending = Mixture2D([[0, 0], [2, 1.8]], [(np.sqrt(0.5), 1, 0.9), (1, 1, 0.8)], [0.6, 0.4], xmin=-1,
                                 label='bending')

        self.hammer = Mixture2D([[0, 0], [1, 1.8]], [(np.sqrt(0.5), 1, 0.9), (0.3, 1, -0.7)], [0.5, 0.5],
                                label='hammer')

        cov = make_2D_Cov(np.sqrt(0.5), 1, 0.1)
        self.skew = Mixture2D([[0, 0], [0, 1.2]], [cov, cov / 4], [0.5, 0.5], label='skew')

        cov = make_2D_Cov(np.sqrt(0.5), 1, 0.1)
        self.broadtail = Mixture2D([[0, 0], [0, 0.2]], [cov, cov * 8], [0.9, 0.1], label='broad tail')

        self.rotating = Mixture2D([[0, 0], [0, 0.2]], [(1, 1, 0.5), (2, 2, -0.5)], [0.6, 0.4], label='rotating')

        self.tight = Mixture2D([[0, 0], [2.5, 3.5]], [(1, 1, 0.99), (1, 1.5, 0.98)], [0.6, 0.4], label='tight')

        self.cut_correlated = Gaussian2D([0, 0], (0.7, 1, 0.95), ymin=0.3, xmax=1.2, label='cut correlated')

        self.flat = Gaussian2D([0, 0], (1, 2, 0), ymin=-1, ymax=2.1, xmin=-1, xmax=0.2, label='flat')

        self.shape_set = [self.gauss, self.bending, self.hammer, self.skew, self.broadtail, self.rotating, self.tight,
                          self.cut_correlated, self.flat]

        self.cut_gaussians = self.cutGaussians((0.7, 1, 0.3))

        # these examples are from Wand and Jones 93
        self.bimodal = []
        self.bimodal.append(
            Mixture2D([[-1, 0], [1, 0]], [(2. / 3, 2. / 3, 0), (2. / 3, 2. / 3, 0)], label='bimodal WJ1'))
        self.bimodal.append(
            Mixture2D([[-3. / 2, 0], [3. / 2, 0]], [(1. / 4, 1, 0), (1. / 4, 1, 0)], label='bimodal WJ2'))
        self.bimodal.append(
            Mixture2D([[-1, 1], [1, -1]], [(2. / 3, 2. / 3, 3. / 5), (2. / 3, 2. / 3, 3. / 5)], label='bimodal WJ3'))
        self.bimodal.append(
            Mixture2D([[1, -1], [-1, 1]], [(2. / 3, 2. / 3, 7. / 10), (2. / 3, 2. / 3, 0)], label='bimodal WJ4'))

        self.trimodal = []
        self.trimodal.append(Mixture2D([[-6. / 5, 6. / 5], [6. / 5, -6. / 5], [0, 0]],
                                       [(3. / 5, 3. / 5, 3. / 10), (3. / 5, 3. / 5, -3. / 5), (0.25, 0.25, 0.2)],
                                       weights=[9, 9, 2], label='trimodal WJ1'))
        self.trimodal.append(Mixture2D([[-6. / 5, 0], [6. / 5, 0], [0, 0]],
                                       [(3. / 5, 3. / 5, 0.7), (3. / 5, 3. / 5, 0.7), (0.25, 0.25, -0.7)],
                                       label='trimodal WJ2'))
        self.trimodal.append(Mixture2D([[-1, 0], [1, 2 * np.sqrt(3) / 3], [1, -2 * np.sqrt(3) / 3]],
                                       [(0.6, 0.7, 0.6), (0.6, 0.7, 0), (0.4, 0.7, 0)], weights=[3, 3, 1],
                                       label='trimodal WJ3'))

        self.quadrimodal = []
        self.quadrimodal.append(Mixture2D([[-1, 1], [-1, -1], [1, -1], [1, 1]],
                                          [(2. / 3, 2. / 3, 2. / 5), (2. / 3, 2. / 3, 3. / 5), (2. / 3, 2. / 3, -0.7),
                                           (2. / 3, 2. / 3, -0.5)],
                                          weights=[1, 3, 1, 3], label='quadrimodal'))

        self.all = self.shape_set + self.bimodal + self.trimodal + self.quadrimodal + self.cut_gaussians

    def cutGaussians(self, cov, cut_x=(-2, -1, -0.5, 0, 1, 1.5, 2)):
        return [Gaussian2D([0, 0], cov, xmin=cut, label=r'Gaussian [$x>%s$]' % cut) for cut in cut_x]

    def distributions(self):
        return self.all


def plot_compare_method(ax, prob, colors=('k',), sims=100, nsamp=default_nsamp,
                        scalings=(0.3, 0.5, 0.7, 0.9, 1, 1.1, 1.3, 1.5, 1.7), test_settings=(None,), linestyles=('-',)):
    # compare Parzen estimator with higher order
    print(prob.label, ', size = ', nsamp)
    if len(colors) == 1:
        colors = colors * len(scalings)
    if len(linestyles) == 1:
        linestyles = linestyles * len(scalings)
    miselist = np.empty((len(scalings), len(test_settings)))
    for i, (settings, ls, color) in enumerate(zip(test_settings, linestyles, colors)):
        if prob.dim == 1:
            scales, MISEs = get1DMises(prob, nsamp=nsamp, scales=scalings, nsim=sims, settings=settings)
        else:
            scales, MISEs = get2DMises(prob, nsamp=nsamp, scales=scalings, nsim=sims, settings=settings)
        ax.plot(scales, MISEs, ls=ls, color=color)
        miselist[:, i] = MISEs
    for i, scale in enumerate(scalings):
        print(scale, miselist[i, :])
    ax.set_yscale('log')
    ax.set_xlim([scalings[0], scalings[-1]])


def plot_compare_probs_methods(ax, probs, colors=plt.rcParams["axes.prop_cycle"].by_key()["color"], **kwargs):
    for prob, col in zip(probs, colors):
        plot_compare_method(ax, prob, col, **kwargs)


def compare_method_nsims(g, probs, sizes=(1000, 10000), **kwargs):
    g.make_figure(len(sizes))
    for i, size in enumerate(sizes):
        ax = g._subplot_number(i)
        plot_compare_probs_methods(ax, probs, nsmap=size, **kwargs)


def compare_method(probs, nx=2, fname='', **kwargs):
    ny = (len(probs) - 1) // nx + 1
    # noinspection PyTypeChecker
    fig, axs = plt.subplots(ny, nx, sharex=True, sharey=True, squeeze=False, figsize=(nx * 3, ny * 3))
    for i, prob in enumerate(probs):
        ax = axs.reshape(-1)[i]
        plot_compare_method(ax, prob, **kwargs)
        ax.text(0.05, 0.06, prob.label, transform=ax.transAxes,
                horizontalalignment='left', bbox=dict(facecolor='floralwhite', alpha=0.9))
        ax.axvline(1, color='gray', ls='--', alpha=0.5)
        if prob.dim == 2:
            if kwargs.get('nsamp') > 15000:
                ax.set_ylim(6e-6, 8e-3)
            elif kwargs.get('nsamp') > 5000:
                ax.set_ylim(2e-4, 5e-2)
        else:
            if kwargs.get('nsamp') > 15000:
                ax.set_ylim(6e-6, 8e-4)
            elif kwargs.get('nsamp') > 5000:
                ax.set_ylim(4e-5, 6e-3)
    plt.subplots_adjust(wspace=0, hspace=0)
    if fname:
        fig.savefig(fname, bbox_inches='tight')


def join_subplots(ax_array):
    for ax in ax_array.reshape(-1):
        if ax is not None:
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
    plt.subplots_adjust(wspace=0, hspace=0)


def run_test_program(plots=('dists_2D', 'dists_1D'), sims=100, nsamp=default_nsamp, mbc=1, bco=1):
    import time

    chains.print_load_details = False
    plt.rc("ytick", direction="in")
    plt.rc("xtick", direction="in")

    test1D = Test1DDistributions()
    test2D = Test2DDistributions()
    test_settings = {'mult_bias_correction_order': mbc, 'boundary_correction_order': bco,
                     'smooth_scale_1D': -1, 'smooth_scale_2D': -1}
    g = get_subplot_plotter(subplot_size=2)

    colors = ['k', 'C0', 'C1', 'C2', 'C3', 'C4']

    if 'ISE_1D' in plots:
        compare_method(test1D.distributions(), nx=3,
                       test_settings=[{'mult_bias_correction_order': 1, 'boundary_correction_order': 1},
                                      {'mult_bias_correction_order': 2, 'boundary_correction_order': 1},
                                      {'mult_bias_correction_order': 0, 'boundary_correction_order': 0},
                                      {'mult_bias_correction_order': 0, 'boundary_correction_order': 1},
                                      {'mult_bias_correction_order': 0, 'boundary_correction_order': 2},
                                      ], colors=colors, linestyles=['-', '-', ':', '-.', '--'],
                       fname='compare_method_1d_N%s.pdf' % nsamp,
                       sims=sims, nsamp=nsamp
                       )

    if 'ISE_2D' in plots:
        compare_method(test2D.distributions(), nx=4,
                       test_settings=[{'mult_bias_correction_order': 1, 'boundary_correction_order': 1},
                                      {'mult_bias_correction_order': 2, 'boundary_correction_order': 1},
                                      {'mult_bias_correction_order': 0, 'boundary_correction_order': 0},
                                      {'mult_bias_correction_order': 0, 'boundary_correction_order': 1},
                                      ], colors=colors, linestyles=['-', '-', ':', '-.', '--'],
                       fname='compare_method_2d_N%s.pdf' % nsamp,
                       sims=sims, nsamp=nsamp
                       )

    if plots is None or 'dists_1D' in plots:
        g.new_plot()
        start = time.time()
        compare1D(g, test1D.distributions(), nsamp=nsamp, settings=test_settings)
        print('1D timing:', time.time() - start)
        join_subplots(g.subplots)
        plt.savefig('test_dists_1D_mbc%s_bco%s_N%s.pdf' % (mbc, bco, nsamp), bbox_inches='tight')

    if plots is None or 'dists_2D' in plots:
        g.new_plot()
        start = time.time()
        compare2D(g, test2D.distributions(), nsamp=nsamp, settings=test_settings)
        print('2D timing:', time.time() - start)
        join_subplots(g.subplots)
        plt.savefig('test_dists_2D_mbc%s_bco%s_N%s.pdf' % (mbc, bco, nsamp), bbox_inches='tight')

    # if False:
    #     print('testing 1D gaussian MISE...')
    #     scales, MISEs = get1DMises(test1D.gauss)
    #     for scale, MISE in zip(scales, MISEs):
    #         print(scale, MISE, np.sqrt(MISE))
    #     print('testing 2D gaussian MISE...')
    #     scales, MISEs = get2DMises(test2D.gauss)
    #     for scale, MISE in zip(scales, MISEs):
    #         print(scale, MISE, np.sqrt(MISE))


if __name__ == "__main__":
    # program to make the 1D and 2D plots in the getdist notes comparison method accuracies
    # To get all four plots use python test_distributions.py --plots ISE_1D ISE_2D dists_1D dists_2D

    # run_test_program(sims=10, nsamp=default_nsamp)
    import argparse

    parser = argparse.ArgumentParser(description='make getdist test plots from test Gaussian mixture distributions')
    parser.add_argument('--sims', type=int, default=100, help='Number of simulations per case')
    parser.add_argument('--nsamp', type=int, default=10000, help='Number of (independent) samples per simulation')
    parser.add_argument('--plots', nargs='*', default=['dists_1D', 'dists_2D', 'ISE_1D', 'ISE_2D'],
                        help='names of plots to make')
    parser.add_argument('--mbc', type=int, default=1, help='baseline mult_bias_correction_order')
    parser.add_argument('--bco', type=int, default=1, help='baseline boundary_correction_order')
    args = parser.parse_args()
    run_test_program(**vars(args))



import os
import numpy as np


class ParamBounds:
    """
    Class for holding list of parameter bounds (e.g. for plotting, or hard priors).
    A limit is None if not specified, denoted by 'N' if read from a string or file

    :ivar names: list of parameter names
    :ivar lower: dict of lower limits, indexed by parameter name
    :ivar upper: dict of upper limits, indexed by parameter name
    """

    def __init__(self, fileName=None):
        """
        :param fileName: optional file name to read from
        """
        self.names = []
        self.lower = {}
        self.upper = {}
        if fileName is not None:
            self.loadFromFile(fileName)

    def loadFromFile(self, fileName):
        self.filenameLoadedFrom = os.path.split(fileName)[1]
        extension = os.path.splitext(fileName)[-1]
        if extension in ('.ranges', '.bounds'):
            with open(fileName, encoding='utf-8-sig') as f:
                for line in f:
                    strings = [text.strip() for text in line.split()]
                    if len(strings) == 3:
                        self.setRange(strings[0], strings[1:])
        elif extension in ('.yaml', '.yml'):
            from getdist.cobaya_interface import get_range, get_info_params, is_parameter_with_range
            info_params = get_info_params(fileName)
            for p, info in info_params.items():
                if is_parameter_with_range(info):
                    self.setRange(p, get_range(info))
        else:
            raise ValueError('ParamBounds must be loaded from .bounds, .ranges or .yaml/.yml file, '
                             'not %s' % fileName)

    def __str__(self):
        s = ''
        for name in self.names:
            valMin = self.getLower(name)
            if valMin is not None:
                lim1 = "%15.7E" % valMin
            else:
                lim1 = "    N"
            valMax = self.getUpper(name)
            if valMax is not None:
                lim2 = "%15.7E" % valMax
            else:
                lim2 = "    N"
            s += "%22s%17s%17s\n" % (name, lim1, lim2)
        return s

    def saveToFile(self, fileName):
        """
        Save to a plain text file

        :param fileName: file name to save to
        """
        with open(fileName, 'w', encoding='utf-8') as f:
            f.write(str(self))

    def _check_name(self, name):
        if not isinstance(name, str):
            raise ValueError('"name" must be a parameter name string not %s: %s' % (type(name), name))

    def setFixed(self, name, value):
        self.setRange(name, (value, value))

    def setRange(self, name, strings):
        self._check_name(name)
        if strings[0] != 'N' and strings[0] is not None and strings[0] != -np.inf:
            self.lower[name] = float(strings[0])
        if strings[1] != 'N' and strings[1] is not None and strings[1] != np.inf:
            self.upper[name] = float(strings[1])
        if name not in self.names:
            self.names.append(name)

    def getUpper(self, name):
        """
        :param name: parameter name
        :return: upper limit, or None if not specified
        """
        self._check_name(name)
        return self.upper.get(name, None)

    def getLower(self, name):
        """
        :param name: parameter name
        :return: lower limit, or None if not specified
        """
        self._check_name(name)
        return self.lower.get(name, None)

    def fixedValue(self, name):
        """
        :param name: parameter name
        :return: if range has zero width return fixed value else return None
        """
        lower = self.lower.get(name, None)
        if lower is not None:
            higher = self.upper.get(name, None)
            if higher is not None:
                if higher == lower:
                    return lower
        return None

    def fixedValueDict(self):
        """
        :return: dictionary of fixed parameter values
        """
        res = {}
        for name in self.names:
            value = self.fixedValue(name)
            if value is not None:
                res[name] = value
        return res



import os
from getdist import plots


# Style that roughly follows the Planck parameter papers; uses latex formatting and sans-serif font.

class PlanckPlotter(plots.GetDistPlotter):
    # common setup for matplotlib
    _style_rc = {'axes.labelsize': 9,
                 'font.size': 8,
                 'legend.fontsize': 8,
                 'xtick.labelsize': 8,
                 'ytick.labelsize': 8,
                 'ytick.major.pad': 4,
                 'xtick.major.pad': 4,
                 'text.usetex': True,
                 'text.latex.preamble': r'\usepackage{%s}' % (
                     os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sfmath').replace(os.sep, '/')),
                 'font.family': 'sans-serif',
                 'font.sans-serif': ['FreeSans', 'Tahoma', 'DejaVu Sans', 'Verdana']}

    def set_default_settings(self):
        s = plots.GetDistPlotSettings()
        s.rc_sizes()
        s.legend_frame = False
        s.figure_legend_frame = False
        s.prob_label = r'$P/P_{\rm max}$'
        s.norm_prob_label = 'Probability density'
        s.prob_y_ticks = True
        s.alpha_filled_add = 0.85
        s.solid_contour_palefactor = 0.6

        s.solid_colors = [('#8CD3F5', '#006FED'), ('#F7BAA6', '#E03424'), ('#D1D1D1', '#A1A1A1'), 'g', 'cadetblue',
                          'olive', 'darkcyan']
        s.axis_marker_lw = 0.6
        s.linewidth_contour = 1
        s.colorbar_axes_fontsize = 8
        # This is how to override default labels for parameters with specified names
        s.param_names_for_labels = os.path.normpath(os.path.join(os.path.dirname(__file__), 'planck.paramnames'))
        self.settings = s

    @classmethod
    def get_single_plotter(cls, **kwargs):
        scaling = kwargs.pop('scaling', None)
        kwargs.pop('rc_sizes', None)
        width_inch = kwargs.pop("width_inch", None) or 3.464
        return super().get_single_plotter(scaling=scaling if scaling is not None else False,
                                          rc_sizes=True, width_inch=width_inch, **kwargs)

    @classmethod
    def get_subplot_plotter(cls, **kwargs):
        scaling = kwargs.pop('scaling', None)
        kwargs.pop('rc_sizes', None)
        return super().get_subplot_plotter(scaling=scaling if scaling is not None else False,
                                           rc_sizes=True, **kwargs)


style_name = 'planck'
plots.add_plotter_style(style_name, PlanckPlotter)



import numpy as np


class CovMat:
    """
    Class holding a covariance matrix for some named parameters

    :ivar matrix: the covariance matrix  (square numpy array)
    :ivar paramNames: list of parameter name strings
    """

    def __init__(self, filename='', matrix=None, paramNames=None):
        """
        :param filename: optionally, a file name to load from

        """
        if not paramNames:
            paramNames = []
        self.matrix = matrix
        self.paramNames = paramNames
        self.size = 0
        if matrix is not None:
            self.size = matrix.shape[0]
        if filename != '':
            self.loadFromFile(filename)

    def paramNameString(self):
        return " ".join(self.paramNames)

    def loadFromFile(self, filename):
        with open(filename, encoding="utf-8-sig") as f:
            first = f.readline().strip()
            if first.startswith('#'):
                self.paramNames = first[1:].split()
                self.size = len(self.paramNames)
            else:
                raise Exception('.covmat must now have parameter names header')
            # noinspection PyTypeChecker
            self.matrix = np.loadtxt(f)

    def saveToFile(self, filename):
        """
        Save the covariance matrix to a text file, with comment header listing the parameter names

        :param filename: name of file to save to (.covmat)
        """
        with open(filename, 'wb') as fout:
            fout.write(('# ' + self.paramNameString() + '\n').encode('UTF-8'))
            np.savetxt(fout, self.matrix, '%15.7E')

    def rescaleParameter(self, name, scale):
        """
        Used to rescale a covariance if a parameter is renormalized

        :param name: parameter name to rescale
        :param scale: value to rescale by
        """
        if name in self.paramNames:
            i = self.paramNames.index(name)
            self.matrix[:, i] = self.matrix[:, i] * scale
            self.matrix[i, :] = self.matrix[i, :] * scale
        else:
            print('Not in covmat: ' + name)

    def mergeCovmatWhereNew(self, cov2):
        params1 = self.paramNames
        params2 = cov2.paramNames

        C = CovMat()
        C.paramNames.extend(params1)

        for param in cov2.paramNames:
            if param not in C.paramNames:
                C.paramNames.append(param)
        l1 = len(params1)
        l2 = len(params2)
        length = len(C.paramNames)

        map1 = dict(list(zip(params1, list(range(0, l1)))))
        map2 = dict(list(zip(params2, list(range(0, l2)))))
        covmap = dict(list(zip(list(range(0, length)), C.paramNames)))

        C.matrix = np.zeros((length, length))
        for i in range(0, length):
            for j in range(0, length):
                if C.paramNames[i] in params1 and C.paramNames[j] in params1:
                    C.matrix[i, j] = self.matrix[map1[covmap[i]], map1[covmap[j]]]
                elif C.paramNames[i] in params2 and C.paramNames[j] in params2:
                    C.matrix[i, j] = cov2.matrix[map2[covmap[i]], map2[covmap[j]]]

        return C

    def correlation(self):
        """
        Get the correlation matrix

        :return: numpy array giving the correlation matrix
        """
        m = self.matrix.copy()
        for i in range(self.size):
            s = np.sqrt(self.matrix[i, i])
            m[i, :] /= s
            m[:, i] /= s
        return m

    def plot(self):
        """
        Plot the correlation matrix as grid of colored squares
        """
        import matplotlib.pyplot as plt

        plt.pcolor(self.correlation())
        plt.colorbar()
        sz = self.size
        plt.yticks(np.arange(0.5, sz + .5), list(range(1, sz + 1)))
        plt.gca().set_yticklabels(self.paramNames)
        plt.xticks(np.arange(0.5, sz + .5), list(range(1, sz + 1)))
        plt.xlim([0, sz])
        plt.ylim([0, sz])
